/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "lql-class.h"
#include "lql-class-private.h"
#include "lql_class_helper.h"

/********************************************************************
 * Non-Virtual API
 ********************************************************************/

/**
 * lql_class_new:
 *
 * Build a new instance of LQLClass. 
 * 
 * Return value: A new LQLClass instance.
 **/
LQLClass*
lql_class_new(void)
{
        return(LQL_CLASS(g_object_new(LQL_TYPE_CLASS, NULL)));
}

/**
 * lql_class_get_id:
 * @self: The LQLClass instance.
 * 
 * Get the ID of this class. In relation to the TC command line options this is the part of the
 * class identifier after the colon.
 * 
 * Return value: The ID of the class
 **/
guint16
lql_class_get_id(LQLClass *self)
{
        guint16 id = 0;
	guint32 tmp = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(-1);
        }
        
        g_object_get(LQL_ELEMENT(self), "handle", &tmp, NULL);

	id = (tmp & 0x0000FFFF);

        return(id);
}

/**
 * lql_class_set_id:
 * @self: 
 * @id: 
 * 
 * Set the ID of this class. In relation to the TC command line options this is the part
 * of the class identifier after the colon.
 * 
 * Return value: 
 **/
gboolean
lql_class_set_id(LQLClass *self,
		 guint16 id)
{
	guint32 tmp = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	g_object_get(LQL_ELEMENT(self), "handle", &tmp, NULL);
	
	tmp = (tmp & 0xFFFF0000) | id;
        
        g_object_set(LQL_ELEMENT(self), "handle", tmp, NULL);
        
        return(TRUE);
}

/**
 * lql_class_get_priority:
 * @self: The LQLClass instance.
 * 
 * Get the priority value for this class.
 * 
 * Return value: The priority value of this class.
 **/
unsigned int
lql_class_get_priority(LQLClass *self)
{
        unsigned int priority = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(-1);
        }
        
        g_object_get(LQL_CLASS(self), "priority", &priority, NULL);

        return(priority);
}

/**
 * lql_class_set_priority:
 * @self: The LQLClass instance.
 * @priority: The priority value to set.
 * 
 * Set the priority value of this class.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_class_set_priority(LQLClass *self,
		       unsigned int priority)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_CLASS(self), "priority", priority, NULL);
        
        return(TRUE);
}

/**
 * lql_class_set_parent:
 * @self: The LQLClass instance.
 * @element: The LQLElement to set as the parent.
 * 
 * Set the parent of this LQLClass to be the passed LQLElement.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_class_set_parent(LQLClass *self,
		     LQLElement *element)
{
	guint32 tmp = 0;

	g_assert(self != NULL);
	g_assert(element != NULL);
	
	if (self->private->dispose_has_run) {
                return(FALSE);
        }

	tmp = lql_element_get_handle(LQL_ELEMENT(element));

	g_object_set(LQL_ELEMENT(self), "parent-handle", tmp, NULL);
	g_object_set(LQL_ELEMENT(self), "handle", (tmp & 0xFFFF0000), NULL);

	return(TRUE);
}

/**
 * lql_class_get_bytes:
 * @self: The LQLClass instance.
 * 
 * Get the number of bytes that have been passed by this Class.
 * 
 * Return value: The number of bytes.
 **/
guint64
lql_class_get_bytes(LQLClass *self)
{
	guint64 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_CLASS(self), "enqueued-bytes", &stat, NULL);

	return(stat);
}

/**
 * lql_class_get_packets:
 * @self: The LQLClass instance.
 * 
 * Get the number of packets that have been passed by this Class.
 * 
 * Return value: The number of packets.
 **/
guint32
lql_class_get_packets(LQLClass *self)
{
	guint32 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_CLASS(self), "enqueued-packets", &stat, NULL);

	return(stat);
}

/**
 * lql_class_get_drops:
 * @self: The LQLClass instance.
 * 
 * Get the number of packets dropped by this Class.
 * 
 * Return value: The number of drops.
 **/
guint32
lql_class_get_drops(LQLClass *self)
{
	guint32 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_CLASS(self), "dropped-packets", &stat, NULL);

	return(stat);
}

/**
 * lql_class_get_overlimits:
 * @self: The LQLClass instance.
 * 
 * Get the number of over limits packets.
 * 
 * Return value: The number of over limits packets.
 **/
guint32
lql_class_get_overlimits(LQLClass *self)
{
	guint32 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_CLASS(self), "overlimits", &stat, NULL);

	return(stat);
}

/**
 * lql_class_get_bps:
 * @self: The LQLClass instance.
 * 
 * Get the current bytes per second rate for this Class.
 * 
 * Return value: The current BPS.
 **/
guint32
lql_class_get_bps(LQLClass *self)
{
	guint32 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_CLASS(self), "bps", &stat, NULL);

	return(stat);
}

/**
 * lql_class_get_pps:
 * @self: The LQLClass instance.
 * 
 * Get the current packets per second rate for this Class.
 * 
 * Return value: The current PPS.
 **/
guint32
lql_class_get_pps(LQLClass *self)
{
	guint32 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_CLASS(self), "pps", &stat, NULL);

	return(stat);
}

/**
 * lql_class_get_qlen:
 * @self: The LQLClass instance.
 * 
 * Get the current queue length for this Class.
 * 
 * Return value: The current queue length.
 **/
guint32
lql_class_get_qlen(LQLClass *self)
{
	guint32 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_CLASS(self), "queue-length", &stat, NULL);

	return(stat);
}

/**
 * lql_class_get_backlog:
 * @self: The LQLClass instance.
 * 
 * Get the current number of backlog packets.
 * 
 * Return value: The number of backlog packets.
 **/
guint32
lql_class_get_backlog(LQLClass *self)
{
	guint32 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_CLASS(self), "backlog", &stat, NULL);

	return(stat);
}

/********************************************************************
 * Virtual API
 ********************************************************************/

/**
 * lql_class_fill_from_nlhdr:
 * @self: The LQLClass object to work with.
 * @hdr: The class information from the kernel.
 * 
 * Fill in the instance variables for this instance of LQLClass
 * to match the network class information in hdr.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_class_fill_from_nlhdr(LQLClass *self,
			  const struct nlmsghdr *hdr)
{
        g_assert(self != NULL);
        g_assert(hdr != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	return(LQL_CLASS_GET_CLASS(self)->fill_from_nlhdr(self, hdr));
}

/**
 * lql_class_add:
 * @self: The LQLClass object to add to the system.
 * @con: The LQLCon connection to use when adding this queueing discipline.
 *
 * Add this queueing discipline to the system.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_class_add(LQLClass *self,
	      LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);
	
        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	return(LQL_CLASS_GET_CLASS(self)->add(self, con));
}

/**
 * lql_class_modify:
 * @self: The LQLClass object to modify.
 * @con: The LQLCon connection to use when adding this class.
 *
 * Modify the queueing discipline in the system to match the current
 * object state.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_class_modify(LQLClass *self,
		 LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);
	
        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	return(LQL_CLASS_GET_CLASS(self)->modify(self, con));
}

/**
 * lql_class_delete:
 * @self: The LQLClass object to delete from the system.
 * @con: The LQLCon connection to use when deleting this queueing discipline.
 *
 * Delete this queueing discipline from the system.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_class_delete(LQLClass *self,
		 LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);
	
        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	return(LQL_CLASS_GET_CLASS(self)->delete(self, con));
}

void
lql_class_print(LQLClass *self)
{
	g_assert(self != NULL);
	
        if (self->private->dispose_has_run) {
                return;
        }

	LQL_CLASS_GET_CLASS(self)->print(self);
}

void
lql_class_printStats(LQLClass *self)
{
	g_assert(self != NULL);
	
        if (self->private->dispose_has_run) {
                return;
        }

	LQL_CLASS_GET_CLASS(self)->printStats(self);
}

gboolean
lql_class_updateStats(LQLClass *self,
		      LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);
	
        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	return(LQL_CLASS_GET_CLASS(self)->updateStats(self, con));	
}

/********************************************************************
 * Object stuff.
 ********************************************************************/
enum {
	LQL_CLASS_PRIORITY = 1,
        LQL_CLASS_BYTES = 2,
	LQL_CLASS_PACKETS = 3,
	LQL_CLASS_DROPS = 4,
	LQL_CLASS_OVERLIMITS = 5,
	LQL_CLASS_BPS = 6,
	LQL_CLASS_PPS = 7,
	LQL_CLASS_QLEN = 8,
	LQL_CLASS_BACKLOG = 9,
};

static void
lql_class_instance_init(GTypeInstance *instance,
			gpointer g_class)
{
        LQLClass *self = (LQLClass *)instance;
        self->private = g_new0(LQLClassPrivate, 1);
        self->private->dispose_has_run = FALSE;
}

static void
lql_class_set_property(GObject *object,
		       guint property_id,
		       const GValue *value,
		       GParamSpec *pspec)
{
        LQLClass *self = (LQLClass *) object;

        switch (property_id) {
	case LQL_CLASS_PRIORITY:
	{
		self->private->priority = g_value_get_uint(value);
		break;
	}
	case LQL_CLASS_BYTES:
	{
                self->private->bytes = g_value_get_int64(value);		
		break;
	}
	case LQL_CLASS_PACKETS:
	{
		self->private->packets = g_value_get_uint(value);
		break;
	}
	case LQL_CLASS_DROPS:
	{
		self->private->drops = g_value_get_uint(value);
		break;
	}
	case LQL_CLASS_OVERLIMITS:
	{
		self->private->overlimits = g_value_get_uint(value);
		break;
	}
	case LQL_CLASS_BPS:
	{
		self->private->bps = g_value_get_uint(value);
		break;
	}
	case LQL_CLASS_PPS:
	{
		self->private->pps = g_value_get_uint(value);
		break;
	}
	case LQL_CLASS_QLEN:
	{
		self->private->qlen = g_value_get_uint(value);
		break;
	}
	case LQL_CLASS_BACKLOG:
	{
		self->private->backlog = g_value_get_uint(value);
		break;
	}
        default:
                g_assert (FALSE);
                break;
        }
}

static void
lql_class_get_property(GObject *object,
		       guint property_id,
		       GValue *value,
		       GParamSpec *pspec)
{
        LQLClass *self = (LQLClass *) object;

        switch (property_id) {
	case LQL_CLASS_PRIORITY:
	{
		g_value_set_uint(value, self->private->priority);
		break;
	}
	case LQL_CLASS_BYTES:
	{
                g_value_set_int64(value, self->private->bytes);
		break;
	}
	case LQL_CLASS_PACKETS:
	{
		g_value_set_uint(value, self->private->packets);
		break;
	}
	case LQL_CLASS_DROPS:
	{
		g_value_set_uint(value, self->private->drops);
		break;
	}
	case LQL_CLASS_OVERLIMITS:
	{
		g_value_set_uint(value, self->private->overlimits);
		break;
	}
	case LQL_CLASS_BPS:
	{
		g_value_set_uint(value, self->private->bps);
		break;
	}
	case LQL_CLASS_PPS:
	{
		g_value_set_uint(value, self->private->pps);
		break;
	}
	case LQL_CLASS_QLEN:
	{
		g_value_set_uint(value, self->private->qlen);
		break;
	}
	case LQL_CLASS_BACKLOG:
	{
		g_value_set_uint(value, self->private->backlog);
		break;
	}
        default:
                g_assert (FALSE);
                break;
        }
}

static GObject *
lql_class_constructor(GType type,
		      guint n_construct_properties,
		      GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                LQLClassClass *klass;
                GObjectClass *parent_class;  
                klass = LQL_CLASS_CLASS(g_type_class_peek(LQL_TYPE_CLASS));
                parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));
                obj = parent_class->constructor(type,
						n_construct_properties,
						construct_properties);
        }
        
        /* do stuff. */

        return obj;
}

static void
lql_class_dispose (GObject *obj)
{
        LQLClass *self = (LQLClass *)obj;

        if (self->private->dispose_has_run) {
                /* If dispose did already run, return. */
                return;
        }
        /* Make sure dispose does not run twice. */
        self->private->dispose_has_run = TRUE;

        /* 
         * In dispose, you are supposed to free all types referenced from this
         * object which might themselves hold a reference to self. Generally,
         * the most simple solution is to unref all members on which you own a 
         * reference.
         */
}

static void
lql_class_finalize (GObject *obj)
{
        LQLClass *self = (LQLClass *)obj;

        /*
         * Complete object destruction.
         */
        
        g_free (self->private);
}

static void
lql_class_class_init (gpointer g_class,
		      gpointer g_class_data)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
        LQLClassClass *klass = LQL_CLASS_CLASS(g_class);
        GParamSpec *pspec = NULL;

        gobject_class->set_property = lql_class_set_property;
        gobject_class->get_property = lql_class_get_property;
        gobject_class->dispose = lql_class_dispose;
        gobject_class->finalize = lql_class_finalize;
        gobject_class->constructor = lql_class_constructor;

	/* Set the virtual methods. */
	klass->fill_from_nlhdr = NULL;
	klass->add = NULL;
	klass->modify = NULL;
	klass->delete = lql_class_helper_delete;
	klass->print = lql_class_helper_print;
	klass->printStats = lql_class_helper_printStats;
	klass->updateStats = lql_class_helper_updateStats;

	/* Setup the properties. */
        pspec = g_param_spec_uint("priority",
				  "Priority of this class",
				  "Set/Get the priority for this class",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_CLASS_PRIORITY,
                                        pspec);
        pspec = g_param_spec_int64("enqueued-bytes",
				   "The count of how many bytes have been enqueued",
				   "Get the cound of how many bytes have been enqueued.",
				   G_MININT64 /* minimum value */,
				   G_MAXINT64 /* maximum value */,
				   0 /* default value */,
				   G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_CLASS_BYTES,
                                        pspec);

        pspec = g_param_spec_uint("enqueued-packets",
				  "The count of how many packets have been enqueued",
				  "Get the count of how many packets have been enqueued.",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_CLASS_PACKETS,
                                        pspec);

        pspec = g_param_spec_uint("dropped-packets",
				  "The count of how many packets have been dropped",
				  "Get the count of how many packets have been dropped.",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_CLASS_DROPS,
                                        pspec);

        pspec = g_param_spec_uint("overlimits",
				  "The count of how many packets were overlimits",
				  "Get the count of how many packets were overlimits.",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_CLASS_OVERLIMITS,
                                        pspec);

        pspec = g_param_spec_uint("bps",
				  "Rate in bytes per second",
				  "Get the rate in bytes per second.",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_CLASS_BPS,
                                        pspec);

        pspec = g_param_spec_uint("pps",
				  "The rate in packets per second",
				  "Get the rate in packets per second.",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_CLASS_PPS,
                                        pspec);

        pspec = g_param_spec_uint("queue-length",
				  "The length of the packet queue.",
				  "Get the length of the packet queue.",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_CLASS_QLEN,
                                        pspec);

        pspec = g_param_spec_uint("backlog",
				  "The number of backlog packets",
				  "Get the number of backlog packets.",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_CLASS_BACKLOG,
                                        pspec);
}

GType lql_class_get_type (void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (LQLClassClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        lql_class_class_init,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (LQLClass),
                        0,      /* n_preallocs */
                        lql_class_instance_init    /* instance_init */
                };
                type = g_type_register_static(LQL_TYPE_ELEMENT,
					      "LQLClass",
					      &info, 0);
        }

        return type;
}
