/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "lql-qdisc-htb.h"
#include "lql-qdisc-htb-private.h"
#include "lql_qdisc_htb_helper.h"

/********************************************************************
 * Non-Virutal API
 ********************************************************************/

/**
 * lql_qdisc_htb_new:
 *
 * Build a new instance of LQLQDiscHTB. 
 * 
 * Return value: A new LQLQDiscHTB instance.
 **/
LQLQDiscHTB*
lql_qdisc_htb_new(void)
{
        return(LQL_QDISC_HTB(g_object_new(LQL_TYPE_QDISC_HTB, NULL)));
}

/**
 * lql_qdisc_htb_get_defaultClass:
 * @self: The LQLQDiscHTB instance.
 * 
 * Get the handle of the class this LQLQDiscHTB will direct packets to by default.
 * 
 * Return value: The default class handle.
 **/
guint32
lql_qdisc_htb_get_defaultClass(LQLQDiscHTB *self)
{
        guint32 defaultClass = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC_HTB(self), "default-class", &defaultClass, NULL);

        return(defaultClass);
}

/**
 * lql_qdisc_htb_set_defaultClass:
 * @self: The LQLQDisc object to work with.
 * @defaultClass: The handle of the class to set as the default class.
 * 
 * Set the default class field of this object.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_qdisc_htb_set_defaultClass(LQLQDiscHTB *self,
			       guint32 defaultClass)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_HTB(self), "default-class", defaultClass, NULL);
        
        return(TRUE);
}

/**
 * lql_qdisc_htb_get_rateToQuantum:
 * @self: The LQLQDiscHTB object to work with.
 * 
 * Get the value of the rateToQuantum field..
 * 
 * Return value: The rateToQuantum.
 **/
guint32
lql_qdisc_htb_get_rateToQuantum(LQLQDiscHTB *self)
{
        guint32 rateToQuantum = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC_HTB(self), "rate-to-quantum", &rateToQuantum, NULL);

        return(rateToQuantum);
}

/**
 * lql_qdisc_htb_set_rateToQuantum:
 * @self: The LQLQDisc object to work with.
 * @rateToQuantum: The rateToQuantum value.
 * 
 * Set the rateToQuantum value for this LQLQDiscHTB.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_qdisc_htb_set_rateToQuantum(LQLQDiscHTB *self,
				guint32 rateToQuantum)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_HTB(self), "rate-to-quantum", rateToQuantum, NULL);
        
        return(TRUE);
}

/**
 * lql_qdisc_htb_get_directPackets:
 * @self: The LQLQDiscHTB instance.
 * 
 * Get the number of packets that have been directly enqueued into this QDisc
 * instead of into a sub-class.
 * 
 * Return value: The number of directly queued packets.
 **/
guint32
lql_qdisc_htb_get_directPackets(LQLQDiscHTB *self)
{
	g_assert(self != NULL);
	
	return(self->private->directPackets);
}

/********************************************************************
 * Virtual API
 ********************************************************************/

/**
 * lql_qdisc_htb_fill_from_nlhdr:
 * @self: The LQLQDisc object to work with.
 * @hdr: The nl packet to get the queue information from.
 *
 * This method takes a nlmsg from the kernel and set this
 * objects instance variables to match it. This is used
 * to create objects that match the ones currently active
 * in the system. This overrides the default LQLQDisc
 * implementation of this method.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_htb_fill_from_nlhdr(LQLQDisc *self,
			      const struct nlmsghdr *hdr)
{
	g_assert(self != NULL);
	g_assert(hdr != NULL);

	return(lql_qdisc_htb_helper_fill_from_nlhdr(LQL_QDISC_HTB(self), hdr));
}

/**
 * lql_qdisc_htb_add:
 * @self: The LQLQDisc object to add to the system.
 * @con: The LQLCon connection to use when adding this queueing discipline.
 *
 * Add this HTB queueing discipline to the system. This overrides
 * the default implementation of this function in LQLQDisc.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_htb_add(LQLQDisc *self,
		  LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);

	return(lql_qdisc_htb_helper_add(LQL_QDISC_HTB(self), con));
}

/**
 * lql_qdisc_htb_modify:
 * @self: The LQLQDisc object to modify.
 * @con: The LQLCon connection to use when adding this qdisc.
 *
 * Modify the HTB queueing discipline in the system to match the current
 * object state. This method overrides the default
 * implementation in LQLQDisc.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_htb_modify(LQLQDisc *self,
		     LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);

	return(lql_qdisc_htb_helper_modify(LQL_QDISC_HTB(self), con));
}

/********************************************************************
 * Object stuff.
 ********************************************************************/
enum {
        LQL_QDISC_HTB_DEFAULTCLASS = 1,
        LQL_QDISC_HTB_RATETOQUANTUM = 2,
};

static void
lql_qdisc_htb_instance_init(GTypeInstance *instance,
			    gpointer g_class)
{
	LQLQDiscHTB *self = (LQLQDiscHTB *)instance;
	self->private = g_new0(LQLQDiscHTBPrivate, 1);
	self->private->dispose_has_run = FALSE;

	self->private->rateToQuantum = 10;
}

static void
lql_qdisc_htb_set_property(GObject *object,
			   guint property_id,
			   const GValue *value,
			   GParamSpec *pspec)
{
        LQLQDiscHTB *self = (LQLQDiscHTB *) object;

        switch (property_id) {
        case LQL_QDISC_HTB_DEFAULTCLASS:
        {
                self->private->defaultClass = g_value_get_uint(value);
                break;
        }
        case LQL_QDISC_HTB_RATETOQUANTUM:
        {
                self->private->rateToQuantum = g_value_get_uint(value);
                break;
        }
        default:
                g_assert(FALSE);
                break;
        }
}

static void
lql_qdisc_htb_get_property(GObject *object,
                           guint property_id,
                           GValue *value,
                           GParamSpec *pspec)
{
        LQLQDiscHTB *self = (LQLQDiscHTB *) object;

        switch (property_id) {
        case LQL_QDISC_HTB_DEFAULTCLASS: 
        {
                g_value_set_uint(value, self->private->defaultClass);
                break;
        }
        case LQL_QDISC_HTB_RATETOQUANTUM: 
        {
                g_value_set_uint(value, self->private->rateToQuantum);
                break;
        }
        default:
                g_assert(FALSE);
                break;
        }
}

static GObject *
lql_qdisc_htb_constructor(GType type,
                          guint n_construct_properties,
                          GObjectConstructParam *construct_properties)
{
        GObject *obj = NULL;

        {
                /* Invoke parent constructor. */
		LQLQDiscHTBClass *klass;
		GObjectClass *parent_class;  
		klass = LQL_QDISC_HTB_CLASS(g_type_class_peek(LQL_TYPE_QDISC_HTB));
		parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));
		obj = parent_class->constructor(type,
						n_construct_properties,
						construct_properties);
        }
        
        /* do stuff. */

        return obj;
}

static void
lql_qdisc_htb_dispose(GObject *obj)
{
        LQLQDiscHTB *self = (LQLQDiscHTB *)obj;

        if (self->private->dispose_has_run) {
                /* If dispose did already run, return. */
                return;
        }
        /* Make sure dispose does not run twice. */
        self->private->dispose_has_run = TRUE;

        /* 
         * In dispose, you are supposed to free all types referenced from this
         * object which might themselves hold a reference to self. Generally,
         * the most simple solution is to unref all members on which you own a 
         * reference.
         */
}

static void
lql_qdisc_htb_finalize(GObject *obj)
{
        LQLQDiscHTB *self = (LQLQDiscHTB *)obj;

        /*
         * Here, complete object destruction.
         */

        g_free (self->private);
}


static void
lql_qdisc_htb_class_init(gpointer g_class,
			 gpointer g_class_data)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
        LQLQDiscClass *klass = LQL_QDISC_CLASS(g_class);
        GParamSpec *pspec = NULL;

        gobject_class->set_property = lql_qdisc_htb_set_property;
        gobject_class->get_property = lql_qdisc_htb_get_property;
        gobject_class->dispose = lql_qdisc_htb_dispose;
        gobject_class->finalize = lql_qdisc_htb_finalize;
        gobject_class->constructor = lql_qdisc_htb_constructor;

	/* Set the virtual methods. */
	klass->fill_from_nlhdr = lql_qdisc_htb_fill_from_nlhdr;
	klass->add = lql_qdisc_htb_add;
	klass->modify = lql_qdisc_htb_modify;
	klass->print = lql_qdisc_htb_helper_print;
	klass->printStats = lql_qdisc_htb_helper_printStats;
	klass->updateStats = lql_qdisc_htb_helper_updateStats;

	/* Setup the properties. */
        pspec = g_param_spec_uint("default-class",
				  "The default class to send packets to",
				  "Set/Get the ID of the class to send traffic to by default",
				  0 /* minimum value */,
				  G_MAXUINT /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_HTB_DEFAULTCLASS,
                                        pspec);

        pspec = g_param_spec_uint("rate-to-quantum",
				  "The rate2quantum for this LQLQDiscHTB",
				  "Set/Get the rate2quantum for this LQLQDiscHTB",
				  0 /* minimum value */,
				  G_MAXUINT /* maximum value */,
				  10 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_HTB_RATETOQUANTUM,
                                        pspec);
}

GType lql_qdisc_htb_get_type (void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (LQLQDiscHTBClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        lql_qdisc_htb_class_init,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (LQLQDiscHTB),
                        0,      /* n_preallocs */
                        lql_qdisc_htb_instance_init    /* instance_init */
                };
                type = g_type_register_static (LQL_TYPE_QDISC,
                                               "LQLQDiscHTB",
                                               &info, 0);
        }
        return type;
}
