/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "lql-qdisc-sfq.h"
#include "lql-qdisc-sfq-private.h"
#include "lql_qdisc_sfq_helper.h"

/********************************************************************
 * Non-Virutal API
 ********************************************************************/

/**
 * lql_qdisc_sfq_new:
 *
 * Build a new instance of LQLQDiscSFQ. 
 * 
 * Return value: A new LQLQDiscSFQ instance.
 **/
LQLQDiscSFQ*
lql_qdisc_sfq_new(void)
{
        return(LQL_QDISC_SFQ(g_object_new(LQL_TYPE_QDISC_SFQ, NULL)));
}

/**
 * lql_qdisc_sfq_get_perturb:
 * @self: The LQLQDiscSFQ object to work with.
 * 
 * Get the perturb value for this LQLQDiscSFQ. The perturb value specifies how often the
 * SFQ QDisc will change it's hashing algorithm.
 * 
 * Return value: The perturb value.
 **/
int
lql_qdisc_sfq_get_perturb(LQLQDiscSFQ *self)
{
        int perturb = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC_SFQ(self), "perturb", &perturb, NULL);

        return(perturb);
}

/**
 * lql_qdisc_sfq_set_perturb:
 * @self: The LQLQDisc object to work with.
 * @perturb: The value to set perturb to.
 * 
 * Set the perturb value for this LQLQDiscSFQ. The perturb value specifies how often the
 * SFQ QDisc will change it's hashing algorithm.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_qdisc_sfq_set_perturb(LQLQDiscSFQ *self,
			  int perturb)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_SFQ(self), "perturb", perturb, NULL);
        
        return(TRUE);
}

/**
 * lql_qdisc_sfq_get_quantum:
 * @self: The LQLQDiscSFQ object to work with.
 * 
 * Get the value of the quantum field.
 * 
 * Return value: The quantum.
 **/
unsigned
lql_qdisc_sfq_get_quantum(LQLQDiscSFQ *self)
{
        unsigned quantum = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC_SFQ(self), "quantum", &quantum, NULL);

        return(quantum);
}

/**
 * lql_qdisc_sfq_set_quantum:
 * @self: The LQLQDisc object to work with.
 * @quantum: The quantum value.
 * 
 * Set the quantum for this LQLQDiscSFQ.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_qdisc_sfq_set_quantum(LQLQDiscSFQ *self,
			  unsigned int quantum)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_SFQ(self), "quantum", quantum, NULL);
        
        return(TRUE);
}

/********************************************************************
 * Virtual API
 ********************************************************************/

/**
 * lql_qdisc_sfq_fill_from_nlhdr:
 * @self: The LQLQDisc object to work with.
 * @hdr: The nl packet to get the queue information from.
 *
 * This method takes a nlmsg from the kernel and set this
 * objects instance variables to match it. This is used
 * to create objects that match the ones currently active
 * in the system. This overrides the default LQLQDisc
 * implementation of this method.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_sfq_fill_from_nlhdr(LQLQDisc *self,
			      const struct nlmsghdr *hdr)
{
	g_assert(self != NULL);
	g_assert(hdr != NULL);

	return(lql_qdisc_sfq_helper_fill_from_nlhdr(LQL_QDISC_SFQ(self), hdr));
}

/**
 * lql_qdisc_sfq_add:
 * @self: The LQLQDisc object to add to the system.
 * @con: The LQLCon connection to use when adding this queueing discipline.
 *
 * Add this SFQ queueing discipline to the system. This overrides
 * the default implementation of this function in LQLQDisc.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_sfq_add(LQLQDisc *self,
		  LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);

	return(lql_qdisc_sfq_helper_add(LQL_QDISC_SFQ(self), con));
}

/**
 * lql_qdisc_sfq_modify:
 * @self: The LQLQDisc object to modify.
 * @con: The LQLCon connection to use when adding this qdisc.
 *
 * Modify the SFQ queueing discipline in the system to match the current
 * object state. This method overrides the default
 * implementation in LQLQDisc.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_sfq_modify(LQLQDisc *self,
		     LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);

	return(lql_qdisc_sfq_helper_modify(LQL_QDISC_SFQ(self), con));
}

/********************************************************************
 * Object stuff.
 ********************************************************************/
enum {
        LQL_QDISC_SFQ_PERTURB = 1,
        LQL_QDISC_SFQ_QUANTUM = 2,
};

static void
lql_qdisc_sfq_instance_init(GTypeInstance *instance,
			    gpointer g_class)
{
	LQLQDiscSFQ *self = (LQLQDiscSFQ *)instance;
	self->private = g_new0(LQLQDiscSFQPrivate, 1);
	self->private->dispose_has_run = FALSE;

	self->private->perturb = 10;
	self->private->quantum = 10;
}

static void
lql_qdisc_sfq_set_property(GObject *object,
			   guint property_id,
			   const GValue *value,
			   GParamSpec *pspec)
{
        LQLQDiscSFQ *self = (LQLQDiscSFQ *) object;

        switch (property_id) {
        case LQL_QDISC_SFQ_PERTURB:
        {
                self->private->perturb = g_value_get_int(value);
                break;
        }
        case LQL_QDISC_SFQ_QUANTUM:
        {
                self->private->quantum = g_value_get_uint(value);
                break;
        }
        default:
                g_assert(FALSE);
                break;
        }
}

static void
lql_qdisc_sfq_get_property(GObject *object,
                           guint property_id,
                           GValue *value,
                           GParamSpec *pspec)
{
        LQLQDiscSFQ *self = (LQLQDiscSFQ *) object;

        switch (property_id) {
        case LQL_QDISC_SFQ_PERTURB: 
        {
                g_value_set_int(value, self->private->perturb);
                break;
        }
        case LQL_QDISC_SFQ_QUANTUM: 
        {
                g_value_set_uint(value, self->private->quantum);
                break;
        }
        default:
                g_assert(FALSE);
                break;
        }
}

static GObject *
lql_qdisc_sfq_constructor(GType type,
                          guint n_construct_properties,
                          GObjectConstructParam *construct_properties)
{
        GObject *obj = NULL;

        {
                /* Invoke parent constructor. */
		LQLQDiscSFQClass *klass;
		GObjectClass *parent_class;  
		klass = LQL_QDISC_SFQ_CLASS(g_type_class_peek(LQL_TYPE_QDISC_SFQ));
		parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));
		obj = parent_class->constructor(type,
						n_construct_properties,
						construct_properties);
        }
        
        /* do stuff. */

        return obj;
}

static void
lql_qdisc_sfq_dispose(GObject *obj)
{
        LQLQDiscSFQ *self = (LQLQDiscSFQ *)obj;

        if (self->private->dispose_has_run) {
                /* If dispose did already run, return. */
                return;
        }
        /* Make sure dispose does not run twice. */
        self->private->dispose_has_run = TRUE;

        /* 
         * In dispose, you are supposed to free all types referenced from this
         * object which might themselves hold a reference to self. Generally,
         * the most simple solution is to unref all members on which you own a 
         * reference.
         */
}

static void
lql_qdisc_sfq_finalize(GObject *obj)
{
        LQLQDiscSFQ *self = (LQLQDiscSFQ *)obj;

        /*
         * Here, complete object destruction.
         */
        
        g_free (self->private);
}


static void
lql_qdisc_sfq_class_init(gpointer g_class,
			 gpointer g_class_data)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
        LQLQDiscClass *klass = LQL_QDISC_CLASS(g_class);
        GParamSpec *pspec = NULL;

        gobject_class->set_property = lql_qdisc_sfq_set_property;
        gobject_class->get_property = lql_qdisc_sfq_get_property;
        gobject_class->dispose = lql_qdisc_sfq_dispose;
        gobject_class->finalize = lql_qdisc_sfq_finalize;
        gobject_class->constructor = lql_qdisc_sfq_constructor;

	/* Set the virtual methods. */
	klass->fill_from_nlhdr = lql_qdisc_sfq_fill_from_nlhdr;
	klass->add = lql_qdisc_sfq_add;
	klass->modify = lql_qdisc_sfq_modify;
	klass->print = lql_qdisc_sfq_helper_print;

	/* Setup the properties. */
        pspec = g_param_spec_int("perturb",
				 "How often to change the hashing algorithm",
				 "Set/Get how often to this QDisc will change it's hashing"
				 " algorithm",
				 0 /* minimum value */,
				 G_MAXINT /* maximum value */,
				 10 /* default value */,
				 G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_SFQ_PERTURB,
                                        pspec);

        pspec = g_param_spec_uint("quantum",
				  "The quantum",
				  "Set/Get the quantum",
				  0 /* minimum value */,
				  G_MAXUINT /* maximum value */,
				  10 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_SFQ_QUANTUM,
                                        pspec);
}

GType lql_qdisc_sfq_get_type (void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (LQLQDiscSFQClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        lql_qdisc_sfq_class_init,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (LQLQDiscSFQ),
                        0,      /* n_preallocs */
                        lql_qdisc_sfq_instance_init    /* instance_init */
                };
                type = g_type_register_static (LQL_TYPE_QDISC,
                                               "LQLQDiscSFQ",
                                               &info, 0);
        }
        return type;
}
