/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <unistd.h>
#include <string.h>

#include "debug_macros.h"
#include "nl.h"
#include "lql-con.h"
#include "lql-qdisc.h"
#include "lql-qdisc-htb.h"
#include "lql-qdisc-htb-private.h"

gboolean
lql_qdisc_htb_helper_fill_from_nlhdr(LQLQDiscHTB *qdisc,
				     const struct nlmsghdr *hdr)
{
	struct tc_htb_glob *opt = NULL;

	g_assert(qdisc != NULL);
	g_assert(hdr != NULL);
	
	lql_element_set_parentHandle(LQL_ELEMENT(qdisc), nl_tcmsg_get_parent(hdr));
	lql_element_set_handle(LQL_ELEMENT(qdisc), nl_tcmsg_get_handle(hdr));
	lql_element_set_ifIndex(LQL_ELEMENT(qdisc), nl_tcmsg_get_ifindex(hdr));
	
	opt = (struct tc_htb_glob *)nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_HTB_INIT);
	if (opt == NULL) {
		g_printerr("Could not get HTB Init\n");
		return(FALSE);
	}

	lql_qdisc_htb_set_defaultClass(LQL_QDISC_HTB(qdisc), opt->defcls);
	lql_qdisc_htb_set_rateToQuantum(LQL_QDISC_HTB(qdisc), opt->rate2quantum);

	return(TRUE);
}

static gboolean
lql_qdisc_htb_helper_do(LQLQDiscHTB *qdisc,
			LQLCon *con,
			int flags)
{
	TcPacket *pkt = NULL;
	struct nlmsghdr *hdr = NULL;
	struct tc_htb_glob *opt = NULL;
	gchar *kind = "htb";

	g_assert(con != NULL);
	g_assert(qdisc != NULL);

	pkt = nl_tcpacket_new();
	hdr = nl_tcpacket_get_nlmsg(pkt);
	
	nl_nlmsg_set_type(hdr, RTM_NEWQDISC);
	nl_nlmsg_set_flags(hdr, NLM_F_REQUEST | NLM_F_ACK | flags);
	nl_nlmsg_set_pid(hdr, getpid());
	nl_nlmsg_set_seq(hdr, lql_con_get_inc_sequence(LQL_CON(con)));

	nl_tcmsg_set_family(hdr, AF_UNSPEC);
	nl_tcmsg_set_ifindex(hdr, lql_element_get_ifIndex(LQL_ELEMENT(qdisc)));
	nl_tcmsg_set_parent(hdr, lql_element_get_parentHandle(LQL_ELEMENT(qdisc)));
	nl_tcmsg_set_handle(hdr, lql_element_get_handle(LQL_ELEMENT(qdisc)));

	opt = g_malloc0(sizeof(struct tc_htb_glob));
	opt->version = 3;
	opt->rate2quantum = lql_qdisc_htb_get_rateToQuantum(LQL_QDISC_HTB(qdisc));
	opt->defcls = lql_qdisc_htb_get_defaultClass(LQL_QDISC_HTB(qdisc));

	if (!nl_tcpacket_add_rtattr(pkt, TCA_KIND, strlen(kind) + 1,
				    kind)) {
		debug_print(5, "Could not add TCA_KIND");
	}
	if (!nl_tcpacket_add_rtattr(pkt, TCA_OPTIONS, 0, NULL)) {
		debug_print(5, "Could not add TCA_OPTIONS");
	}
	if (!nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_HTB_INIT,
					NLMSG_ALIGN(sizeof(struct tc_htb_glob)), opt)) {
		debug_print(5, "Could not add TCA_HTB_INIT");
	}

	if (!nl_tcpacket_do_command(pkt, lql_con_get_nlcon(LQL_CON(con)))) {
		return(FALSE);
	}

	g_free(opt);
	g_free(pkt);

	return(TRUE);
}

gboolean
lql_qdisc_htb_helper_add(LQLQDiscHTB *qdisc,
			 LQLCon *con)
{
	g_assert(qdisc != NULL);
	g_assert(con != NULL);

	return(lql_qdisc_htb_helper_do(qdisc, con, NLM_F_CREATE | NLM_F_EXCL));
}

gboolean
lql_qdisc_htb_helper_modify(LQLQDiscHTB *qdisc,
			    LQLCon *con)
{
	g_assert(qdisc != NULL);
	g_assert(con != NULL);

	return(lql_qdisc_htb_helper_do(qdisc, con, 0));
}

void
lql_qdisc_htb_helper_print(LQLQDisc *self)
{
	LQLQDiscHTBClass *klass = NULL;
	LQLQDiscClass *parentKlass = NULL;
	LQLQDiscHTB *htb = NULL;

	g_assert(self != NULL);
	g_assert(LQL_IS_QDISC_HTB(self));

	htb = LQL_QDISC_HTB(self);

        if (htb->private->dispose_has_run) {
                return;
        }

	klass = LQL_QDISC_HTB_GET_CLASS(self);
	parentKlass = g_type_class_peek_parent(klass);
	
	parentKlass->print(self);

	g_print("Type: HTB\n");
	g_print("HTB Default class: %u\n", lql_qdisc_htb_get_defaultClass(htb));
	g_print("HTB Rate to quantum: %u\n", lql_qdisc_htb_get_rateToQuantum(htb));
}

void
lql_qdisc_htb_helper_printStats(LQLQDisc *self)
{
	LQLQDiscHTBClass *klass = NULL;
	LQLQDiscClass *parentKlass = NULL;
	LQLQDiscHTB *htb = NULL;

	g_assert(self != NULL);
	g_assert(LQL_IS_QDISC_HTB(self));

	htb = LQL_QDISC_HTB(self);

        if (htb->private->dispose_has_run) {
                return;
        }

	klass = LQL_QDISC_HTB_GET_CLASS(self);
	parentKlass = g_type_class_peek_parent(klass);
	
	parentKlass->printStats(self);

	g_print("HTB direct packets: %u\n", lql_qdisc_htb_get_directPackets(htb));
}

/* static gboolean */
/* lql_qdisc_htb_helper_fill_xstats(LQLQDiscHTB *self, */
/* 				 const struct nlmsghdr *hdr) */
/* { */
/* 	const struct tc_htb_glob *glob = NULL; */

/* 	g_assert(self != NULL); */
/* 	g_assert(hdr != NULL); */

/* 	glob = nl_tcmsg_get_rta_val(hdr, TCA_HTB_INIT); */
/* 	if (glob == NULL) { */
/* 		return(FALSE); */
/* 	} */

/* 	/\* Filter for the QDisc we are looking for since there doesn't appear */
/* 	   to be a way to tell the kernel to only return a single QDisc. *\/ */
/*  	if (nl_tcmsg_get_handle(hdr) == */
/* 	    lql_element_get_handle(LQL_ELEMENT(self))) { */
/* 		self->private->directPackets = glob->direct_pkts; */
/* 	} */

/* 	return(TRUE); */
/* } */

gboolean
lql_qdisc_htb_helper_updateStats(LQLQDisc *self,
				 LQLCon *con)
{
	LQLQDiscHTBClass *klass = NULL;
	LQLQDiscClass *parentKlass = NULL;
	LQLQDiscHTB *htb = NULL;

	g_assert(self != NULL);
	g_assert(LQL_IS_QDISC_HTB(self));

	htb = LQL_QDISC_HTB(self);

        if (htb->private->dispose_has_run) {
                return(FALSE);
        }

	klass = LQL_QDISC_HTB_GET_CLASS(self);
	parentKlass = g_type_class_peek_parent(klass);
	
	return(parentKlass->updateStats(self, con));
}
