/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <asm/types.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <linux/netlink.h>
#include <linux/rtnetlink.h>

#include <string.h>
#include <glib.h>

#include "debug_macros.h"
#include "nl.h"

static gboolean nl_nlmsg_add_rtattr(struct nlmsghdr *hdr, unsigned short type, unsigned short alen, void *data);
static gboolean nl_nlmsg_add_rtattr_sub(struct nlmsghdr *hdr, unsigned short parentType, unsigned short type, unsigned short alen, void *data, int hdrSize);

NLCon *
nl_open(unsigned groups)
{
	NLCon *con = NULL;
	int ret;

	con = g_new0(NLCon, 1);

	con->sock = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE);
	if (con->sock == -1) {
		g_printerr("nl_init(): Could not create socket.\n");
		return(NULL);
	}

	con->addr.nl_family = AF_NETLINK;
	con->addr.nl_pid = 0;
	con->addr.nl_groups = groups;

	ret = bind(con->sock, (struct sockaddr *)&con->addr, sizeof(con->addr));
	if (ret == -1) {
		g_printerr("nl_init(): Could not bind to socket.\n");
		return(NULL);
	}

	return(con);
}

gboolean
nl_send(const NLCon *con,
	const void *buffer,
	const size_t len)
{
	int ret;

	g_assert(con != NULL);
	g_assert(buffer != NULL);
	g_assert(len > 0);
	
	ret = send(con->sock, buffer, len, 0);
	if (ret != len) {
		g_printerr("nl_send(): Could not write to socket.\n");
		return(FALSE);
	}

	return(TRUE);
}

size_t
nl_read(NLCon *con,
	void *buf,
	int len)
{
	int ret = 0;

	g_assert(con != NULL);
	g_assert(buf != NULL);
	g_assert(len > 0);

	ret = recv(con->sock, buf, len, 0);
	
	return(ret);
}

gboolean
nl_close(NLCon *con)
{
	g_assert(con != NULL);

	close(con->sock);

	g_free(con);
	
	return(TRUE);
}

__u32
nl_nlmsg_get_length(const struct nlmsghdr *nlmsg)
{
	g_assert(nlmsg != NULL);
	
	return(nlmsg->nlmsg_len);
}

__u16
nl_nlmsg_get_type(const struct nlmsghdr *nlmsg)
{
	g_assert(nlmsg != NULL);
	
	return(nlmsg->nlmsg_type);
}

__u16
nl_nlmsg_get_flags(const struct nlmsghdr *nlmsg)
{
	g_assert(nlmsg != NULL);
	
	return(nlmsg->nlmsg_flags);
}

__u32
nl_nlmsg_get_pid(const struct nlmsghdr *nlmsg)
{
	g_assert(nlmsg != NULL);
	
	return(nlmsg->nlmsg_pid);
}

__u32
nl_nlmsg_get_seq(const struct nlmsghdr *nlmsg)
{
	g_assert(nlmsg != NULL);
	
	return(nlmsg->nlmsg_seq);
}

gboolean
nl_nlmsg_set_length(struct nlmsghdr *nlmsg, __u32 length)
{
	g_assert(nlmsg != NULL);
	
	nlmsg->nlmsg_len = length;
	
	return(TRUE);
}

gboolean
nl_nlmsg_set_type(struct nlmsghdr *nlmsg, __u16 type)
{
	g_assert(nlmsg != NULL);
	
	nlmsg->nlmsg_type = type;
	
	return(TRUE);
}

gboolean
nl_nlmsg_set_flags(struct nlmsghdr *nlmsg, __u16 flags)
{
	g_assert(nlmsg != NULL);
	
	nlmsg->nlmsg_flags = flags;
	
	return(TRUE);
}

gboolean
nl_nlmsg_set_pid(struct nlmsghdr *nlmsg, __u32 pid)
{
	g_assert(nlmsg != NULL);
	
	nlmsg->nlmsg_pid = pid;
	
	return(TRUE);
}

gboolean
nl_nlmsg_set_seq(struct nlmsghdr *nlmsg, __u32 seq)
{
	g_assert(nlmsg != NULL);
	
	nlmsg->nlmsg_seq = seq;
	
	return(TRUE);
}

gboolean
nl_nlmsg_is_done(const struct nlmsghdr *nlmsg)
{
	g_assert(nlmsg != NULL);

	if(nlmsg->nlmsg_type == NLMSG_DONE) {
		return(TRUE);
	}
	
	return(FALSE);
}

gboolean
nl_nlmsg_is_error(const struct nlmsghdr *nlmsg)
{
	g_assert(nlmsg != NULL);

	if(nlmsg->nlmsg_type == NLMSG_ERROR) {
		return(TRUE);
	}
	
	return(FALSE);
}

gboolean
nl_nlmsg_is_multi(const struct nlmsghdr *nlmsg)
{
	g_assert(nlmsg != NULL);

	if (nlmsg->nlmsg_flags & NLM_F_MULTI) {
		return(TRUE);
	}

	return(FALSE);
}

const struct nlmsghdr*
nl_get_nlmsghdr(const void *buf,
		const int len,
		const int num)
{
	int i = 0;
	struct nlmsghdr *hdr = NULL;

	g_assert(buf != NULL);
	g_assert(len > 0);
	g_assert(num >= 0);

	hdr = (struct nlmsghdr *) buf;

	for (i = 1; i <= num; i++) {
		if (((void *)hdr + nl_nlmsg_get_length(hdr)) >= buf + len) {
			return(NULL);
		}

		hdr = (struct nlmsghdr *) ((void *)hdr + nl_nlmsg_get_length(hdr));
	}

	return(hdr);
}

gboolean
nl_ifinfomsg_verify(const struct nlmsghdr *nlmsg)
{
	g_assert(nlmsg != NULL);

	if ((nlmsg->nlmsg_type == RTM_NEWLINK) ||
	    (nlmsg->nlmsg_type == RTM_GETLINK)) {
		return(TRUE);
	}
	
	return(FALSE);
}

const struct ifinfomsg*
nl_nlmsg_get_ifinfomsg_const(const struct nlmsghdr *nlmsg)
{
	struct ifinfomsg *ifinfo = NULL;

	g_assert(nlmsg != NULL);
	
	if (!nl_ifinfomsg_verify(nlmsg)) {
		return(NULL);
	}

	ifinfo = NLMSG_DATA(nlmsg);

	return(ifinfo);
}

struct ifinfomsg*
nl_nlmsg_get_ifinfomsg(struct nlmsghdr *nlmsg)
{
	struct ifinfomsg *ifinfo = NULL;

	g_assert(nlmsg != NULL);
	
	if (!nl_ifinfomsg_verify(nlmsg)) {
		return(NULL);
	}

	ifinfo = NLMSG_DATA(nlmsg);

	return(ifinfo);
}

const struct tcmsg*
nl_nlmsg_get_tcmsg_const(const struct nlmsghdr *nlmsg)
{
	struct tcmsg *tc = NULL;

	g_assert(nlmsg != NULL);
	
	if (!nl_tcmsg_verify(nlmsg)) {
		return(NULL);
	}

	tc = NLMSG_DATA(nlmsg);

	return(tc);
}

struct tcmsg*
nl_nlmsg_get_tcmsg(struct nlmsghdr *nlmsg)
{
	struct tcmsg *tc = NULL;

	g_assert(nlmsg != NULL);
	
	if (!nl_tcmsg_verify(nlmsg)) {
		return(NULL);
	}

	tc = NLMSG_DATA(nlmsg);

	return(tc);
}

unsigned char
nl_ifinfo_get_family(const struct nlmsghdr *nlmsg)
{
	const struct ifinfomsg *ifinfo = NULL;

	g_assert(nlmsg != NULL);

	ifinfo = nl_nlmsg_get_ifinfomsg_const(nlmsg);
	if (ifinfo == NULL) {
		// NLMSG does not contain a ifinfomsg.
		return(-1);
	}
	
	return(ifinfo->ifi_family);
}

unsigned short
nl_ifinfo_get_type(const struct nlmsghdr *nlmsg)
{
	const struct ifinfomsg *ifinfo = NULL;

	g_assert(nlmsg != NULL);

	ifinfo = nl_nlmsg_get_ifinfomsg_const(nlmsg);
	if (ifinfo == NULL) {
		// NLMSG does not contain a ifinfomsg.
		return(-1);
	}

	return(ifinfo->ifi_type);
}

int
nl_ifinfo_get_index(const struct nlmsghdr *nlmsg)
{
	const struct ifinfomsg *ifinfo = NULL;

	g_assert(nlmsg != NULL);

	ifinfo = nl_nlmsg_get_ifinfomsg_const(nlmsg);
	if (ifinfo == NULL) {
		// NLMSG does not contain a ifinfomsg.
		return(-1);
	}

	return(ifinfo->ifi_index);
}

unsigned
nl_ifinfo_get_change(const struct nlmsghdr *nlmsg)
{
	const struct ifinfomsg *ifinfo = NULL;

	g_assert(nlmsg != NULL);

	ifinfo = nl_nlmsg_get_ifinfomsg_const(nlmsg);
	if (ifinfo == NULL) {
		// NLMSG does not contain a ifinfomsg.
		return(-1);
	}

	return(ifinfo->ifi_change);
}

gboolean
nl_ifinfo_set_family(struct nlmsghdr *nlmsg,
		     unsigned char family)
{
	struct ifinfomsg *ifinfo = NULL;

	g_assert(nlmsg != NULL);

	ifinfo = nl_nlmsg_get_ifinfomsg(nlmsg);
	if (ifinfo == NULL) {
		// NLMSG does not contain a ifinfomsg.
		return(FALSE);
	}

	ifinfo->ifi_family = family;

	return(TRUE);
}

gboolean
nl_ifinfo_set_type(struct nlmsghdr *nlmsg,
		   unsigned short type)
{
	struct ifinfomsg *ifinfo = NULL;

	g_assert(nlmsg != NULL);

	ifinfo = nl_nlmsg_get_ifinfomsg(nlmsg);
	if (ifinfo == NULL) {
		// NLMSG does not contain a ifinfomsg.
		return(FALSE);
	}

	ifinfo->ifi_type = type;

	return(TRUE);
}

gboolean
nl_ifinfo_set_index(struct nlmsghdr *nlmsg,
		    int index)
{
	struct ifinfomsg *ifinfo = NULL;

	g_assert(nlmsg != NULL);

	ifinfo = nl_nlmsg_get_ifinfomsg(nlmsg);
	if (ifinfo == NULL) {
		// NLMSG does not contain a ifinfomsg.
		return(FALSE);
	}

	ifinfo->ifi_index = index;

	return(TRUE);
}

gboolean
nl_ifinfo_set_flags(struct nlmsghdr *nlmsg,
		    unsigned flags)
{
	struct ifinfomsg *ifinfo = NULL;

	g_assert(nlmsg != NULL);

	ifinfo = nl_nlmsg_get_ifinfomsg(nlmsg);
	if (ifinfo == NULL) {
		// NLMSG does not contain a ifinfomsg.
		return(FALSE);
	}

	ifinfo->ifi_flags = flags;

	return(TRUE);
}

gboolean
nl_ifinfo_set_change(struct nlmsghdr *nlmsg,
		     unsigned change)
{
	struct ifinfomsg *ifinfo = NULL;

	g_assert(nlmsg != NULL);

	ifinfo = nl_nlmsg_get_ifinfomsg(nlmsg);
	if (ifinfo == NULL) {
		// NLMSG does not contain a ifinfomsg.
		return(FALSE);
	}

	ifinfo->ifi_change = change;

	return(TRUE);
}

const void*
nl_ifinfo_get_rta_val(const struct nlmsghdr *nlmsg,
		      unsigned short type)
{
	const struct ifinfomsg *ifinfo = NULL;
	struct rtattr *rta = NULL;
	int len = 0;

	g_assert(nlmsg != NULL);

	ifinfo = nl_nlmsg_get_ifinfomsg_const(nlmsg);
	if (ifinfo == NULL) {
		// NLMSG does not contain a ifinfomsg.
		return(NULL);
	}

	rta = IFLA_RTA(ifinfo);

	len = nl_nlmsg_get_length(nlmsg);
	len = len - NLMSG_LENGTH(NLMSG_ALIGN(sizeof(struct ifinfomsg)));
	
	while(RTA_OK(rta, len)) {
		if (rta->rta_type == type) {
			return(RTA_DATA(rta));
		}

		len = len - rta->rta_len;
		
		rta = RTA_NEXT(rta, len);
	}

	return(NULL);
}

IfInfoPacket*
nl_ifinfopacket_new(void) 
{
	IfInfoPacket *pkt = NULL;
	
	pkt = g_malloc0(sizeof(IfInfoPacket));
	
	pkt->hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifinfomsg));
	
	return(pkt);
}

struct nlmsghdr*
nl_ifinfopacket_get_nlmsg(IfInfoPacket *pkt)
{
	struct nlmsghdr *hdr = NULL;

	g_assert(pkt != NULL);

	hdr = (struct nlmsghdr *) pkt;
	
	return(hdr);
}

struct ifinfomsg*
nl_ifinfopacket_get_ifinfomsg(IfInfoPacket *pkt)
{
	struct ifinfomsg *msg = NULL;
	
	g_assert(pkt != NULL);
	
	msg = (struct ifinfomsg *) ((void *)pkt + sizeof(struct nlmsghdr));
	
	return(msg);
}

gboolean
nl_ifinfopacket_add_rtattr(IfInfoPacket *pkt,
			   unsigned short type,
			   unsigned short alen,
			   void *data)
{
	struct nlmsghdr *hdr = NULL;

	g_assert(pkt != NULL);
	if (alen != 0) {
		g_assert(data != NULL);
	}
	

	hdr = nl_ifinfopacket_get_nlmsg(pkt);

	if (!nl_nlmsg_add_rtattr(hdr, type, alen, data)) {
		return(FALSE);
	}
	
	return(TRUE);
}

gboolean
nl_rtattr_set_len(struct rtattr *rta,
		  struct nlmsghdr *hdr)
{
	g_assert(rta != NULL);
	g_assert(hdr != NULL);
	
	rta->rta_len = (((void*)hdr)+NLMSG_ALIGN(hdr->nlmsg_len)) - (void*)rta;

	return(TRUE);
}

/*
 * This function adds a new rtattr to an existing one already in the the
 * TcPacket. Warning: This just appends the new rtattr to the packet and
 * adjusts the nlmsghdr length and parent rta_len. Thus, it can only be used
 * to append a sub rtattr when there is nothing after the parent rtattr.
 * Other sub-rtattrs are OK though.
 * @param pkt TcPacket to work with.
 * @param parentType The rta_type of the parent we are adding this rtattr to.
 * @param type The type of this rtattr.
 * @param alen The length of the data for this rtattr.
 * @param data A pointer to the date for this rtattr.
 * @return TRUE on success, FALSE on failure.
 */
gboolean
nl_tcpacket_add_rtattr_sub(TcPacket *pkt,
			   unsigned short parentType,
			   unsigned short type,
			   unsigned short alen,
			   void *data)
{
	struct nlmsghdr *hdr = NULL;
	int hdrSize = 0;

	g_assert(pkt != NULL);
	if (alen != 0) {
		g_assert(data != NULL);
	}

	hdr = nl_tcpacket_get_nlmsg(pkt);
	hdrSize = NLMSG_ALIGN(sizeof(struct tcmsg));

	if (!nl_nlmsg_add_rtattr_sub(hdr, parentType, type, alen, data, hdrSize)) {
		return(FALSE);
	}
	
	return(TRUE);
}

static gboolean
nl_nlmsg_add_rtattr_sub(struct nlmsghdr *hdr,
			unsigned short parentType,
			unsigned short type,
			unsigned short alen,
			void *data,
			int hdrSize)
{
	int len = 0;
	int hdrLen = 0;
	struct rtattr *rta = NULL;
	struct rtattr *new = NULL;

	g_assert(hdr != NULL);
	if (alen != 0) {
		g_assert(data != NULL);
	}

	len = RTA_LENGTH(RTA_ALIGN(alen));

	hdrLen = NLMSG_ALIGN(nl_nlmsg_get_length(hdr));
	
	if (hdrLen + len > NL_PACKET_DATA_SIZE) {
		debug_print(5, "nl_nlmsg_add_rtattr_sub(): Packet too big.\n");
		return(FALSE);
	}
	
	rta = NLMSG_DATA(hdr) + hdrSize;
	while (rta->rta_type != parentType) {
		rta = (void *)rta + rta->rta_len;

		if ((void *)rta >= (void *)hdr + hdrLen) {
			// Searched past the end. Big error.
			debug_print(5,
				    "nl_nlmsg_add_rtattr_sub(): Could not find parent.\n");
			return(FALSE);
		}
	}

	new = (void *)hdr + hdrLen;
	new->rta_type = type;
	new->rta_len = len;
	memcpy(RTA_DATA(new), data, alen);

	rta->rta_len += len;
	nl_nlmsg_set_length(hdr, hdrLen + len);
	
	return(TRUE);
}

static gboolean
nl_nlmsg_add_rtattr(struct nlmsghdr *hdr,
		    unsigned short type,
		    unsigned short alen,
		    void *data)
{
	int len = 0;
	int hdrLen = 0;
	struct rtattr *rta = NULL;

	g_assert(hdr != NULL);
	if (alen != 0) {
		g_assert(data != NULL);
	}
	
	len = RTA_LENGTH(RTA_ALIGN(alen));

	hdrLen = NLMSG_ALIGN(nl_nlmsg_get_length(hdr));
	
	if (hdrLen + len > NL_PACKET_DATA_SIZE) {
		debug_print(5, "nl_nlmsg_add_rtattr(): Packet too big.\n");
		return(FALSE);
	}

	rta = (struct rtattr *) ((void *)hdr + hdrLen);
	rta->rta_type = type;
	rta->rta_len = len;
	
	memcpy(RTA_DATA(rta), data, alen);
	nl_nlmsg_set_length(hdr, hdrLen + len);

	return(TRUE);
}

TcPacket*
nl_tcpacket_new(void)
{
	TcPacket *pkt = NULL;
	
	pkt = g_malloc0(sizeof(TcPacket));
	
	pkt->hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct tcmsg));
	
	return(pkt);
}

struct nlmsghdr*
nl_tcpacket_get_nlmsg(TcPacket *pkt)
{
	struct nlmsghdr *hdr = NULL;

	g_assert(pkt != NULL);

	hdr = (struct nlmsghdr *) pkt;
	
	return(hdr);
}

struct tcmsg*
nl_tcpacket_get_tcmsg(TcPacket *pkt)
{
	struct tcmsg *msg = NULL;
	
	g_assert(pkt != NULL);
	
	msg = (struct tcmsg *) ((void *)pkt + sizeof(struct nlmsghdr));
	
	return(msg);
}

gboolean
nl_tcpacket_add_rtattr(TcPacket *pkt,
		       unsigned short type,
		       unsigned short alen,
		       void *data)
{
	struct nlmsghdr *hdr = NULL;

	g_assert(pkt != NULL);
	if (alen != 0) {
		g_assert(data != NULL);
	}

	hdr = nl_tcpacket_get_nlmsg(pkt);

	if (!nl_nlmsg_add_rtattr(hdr, type, alen, data)) {
		return(FALSE);
	}
	
	return(TRUE);
}

gboolean
nl_tcpacket_do_command(TcPacket *pkt,
		       NLCon *nlcon)
{
	void *buf = NULL;
	int ret, i;
	const struct nlmsghdr *hdr = NULL;
	
	buf = g_malloc0(NL_PACKET_DATA_SIZE);
	
	g_assert(nlcon != NULL);
	g_assert(pkt != NULL);

	// First send the packet.
	hdr = nl_tcpacket_get_nlmsg(pkt);
	if (!nl_send(nlcon, pkt, nl_nlmsg_get_length(hdr))) {
		g_print("Could not send packet.\n");
	}

	for (;;) {
		ret = nl_read(nlcon, buf, NL_PACKET_DATA_SIZE);
		i = 0;
		
		while ((hdr = nl_get_nlmsghdr(buf, ret, i)) != NULL) {
			
//			g_print("Got a hdr:%i\n", hdr->nlmsg_type);

			if (nl_nlmsg_is_error(hdr)) {
				struct nlmsgerr *err = (struct nlmsgerr*)NLMSG_DATA(hdr);

                                if (hdr->nlmsg_len < NLMSG_LENGTH(sizeof(struct nlmsgerr))) {
                                        g_printerr("ERROR truncated\n");
                                } else {
                                        errno = -err->error;
//					printf("err:%i\n", errno);
//                                      perror("RTNETLINK answers");

					if (errno != 0) {
						perror("RTNETLINK answers");
						return(FALSE);
					}
                                }
			}

			if (!nl_nlmsg_is_multi(hdr)) {
				goto done;
			}

			if (nl_nlmsg_is_done(hdr)) {
				goto done;
			}

			i++;
		}
	}

 done:
	g_free(buf);

	return(TRUE);	
}

gboolean
nl_tcmsg_verify(const struct nlmsghdr *nlmsg)
{
	g_assert(nlmsg != NULL);

	if ((nlmsg->nlmsg_type == RTM_NEWQDISC) ||
	    (nlmsg->nlmsg_type == RTM_DELQDISC) ||
	    (nlmsg->nlmsg_type == RTM_GETQDISC) ||
	    (nlmsg->nlmsg_type == RTM_NEWTCLASS) ||
	    (nlmsg->nlmsg_type == RTM_DELTCLASS) ||
	    (nlmsg->nlmsg_type == RTM_GETTCLASS) ||
	    (nlmsg->nlmsg_type == RTM_NEWTFILTER) ||
	    (nlmsg->nlmsg_type == RTM_DELTFILTER) ||
	    (nlmsg->nlmsg_type == RTM_GETTFILTER)) {
		return(TRUE);
	}
	
	return(FALSE);	
}

unsigned char
nl_tcmsg_get_family(const struct nlmsghdr *nlmsg)
{
	const struct tcmsg *tc = NULL;

	g_assert(nlmsg != NULL);

	tc = nl_nlmsg_get_tcmsg_const(nlmsg);
	if (tc == NULL) {
		// NLMSG does not contain a tcmsg.
		return(-1);
	}
	
	return(tc->tcm_family);
}

int
nl_tcmsg_get_ifindex(const struct nlmsghdr *nlmsg)
{
	const struct tcmsg *tc = NULL;

	g_assert(nlmsg != NULL);

	tc = nl_nlmsg_get_tcmsg_const(nlmsg);
	if (tc == NULL) {
		// NLMSG does not contain a tcmsg.
		return(-1);
	}
	
	return(tc->tcm_ifindex);
}

__u32
nl_tcmsg_get_handle(const struct nlmsghdr *nlmsg)
{
	const struct tcmsg *tc = NULL;

	g_assert(nlmsg != NULL);

	tc = nl_nlmsg_get_tcmsg_const(nlmsg);
	if (tc == NULL) {
		// NLMSG does not contain a tcmsg.
		return(-1);
	}
	
	return(tc->tcm_handle);
}

__u32
nl_tcmsg_get_parent(const struct nlmsghdr *nlmsg)
{
	const struct tcmsg *tc = NULL;

	g_assert(nlmsg != NULL);

	tc = nl_nlmsg_get_tcmsg_const(nlmsg);
	if (tc == NULL) {
		// NLMSG does not contain a tcmsg.
		return(-1);
	}
	
	return(tc->tcm_parent);
}

__u32
nl_tcmsg_get_info(const struct nlmsghdr *nlmsg)
{
	const struct tcmsg *tc = NULL;

	g_assert(nlmsg != NULL);

	tc = nl_nlmsg_get_tcmsg_const(nlmsg);
	if (tc == NULL) {
		// NLMSG does not contain a tcmsg.
		return(-1);
	}
	
	return(tc->tcm_info);
}

gboolean
nl_tcmsg_set_family(struct nlmsghdr *nlmsg,
		    unsigned char family)
{
	struct tcmsg *tc = NULL;

	g_assert(nlmsg != NULL);

	tc = nl_nlmsg_get_tcmsg(nlmsg);
	if (tc == NULL) {
		// NLMSG does not contain a tcmsg.
		debug_print(5, "nl_tcmsg_set_family(): failed.\n");
		return(FALSE);
	}

	tc->tcm_family = family;

	return(TRUE);
}

gboolean
nl_tcmsg_set_ifindex(struct nlmsghdr *nlmsg,
		     int index)
{
	struct tcmsg *tc = NULL;

	g_assert(nlmsg != NULL);

	tc = nl_nlmsg_get_tcmsg(nlmsg);
	if (tc == NULL) {
		// NLMSG does not contain a tcmsg.
		debug_print(5, "nl_tcmsg_set_ifindex(): failed.\n");
		return(FALSE);
	}

	tc->tcm_ifindex = index;

	return(TRUE);
}

gboolean
nl_tcmsg_set_handle(struct nlmsghdr *nlmsg,
		    __u32 handle)
{
	struct tcmsg *tc = NULL;

	g_assert(nlmsg != NULL);

	tc = nl_nlmsg_get_tcmsg(nlmsg);
	if (tc == NULL) {
		// NLMSG does not contain a tcmsg.
		debug_print(5, "nl_tcmsg_set_handle(): failed.\n");
		return(FALSE);
	}

	tc->tcm_handle = handle;

	return(TRUE);
}

gboolean
nl_tcmsg_set_parent(struct nlmsghdr *nlmsg,
		    __u32 parent)
{
	struct tcmsg *tc = NULL;

	g_assert(nlmsg != NULL);

	tc = nl_nlmsg_get_tcmsg(nlmsg);
	if (tc == NULL) {
		// NLMSG does not contain a tcmsg.
		return(FALSE);
	}

	tc->tcm_parent = parent;

	return(TRUE);
}

gboolean
nl_tcmsg_set_info(struct nlmsghdr *nlmsg,
		  __u32 info)
{
	struct tcmsg *tc = NULL;

	g_assert(nlmsg != NULL);

	tc = nl_nlmsg_get_tcmsg(nlmsg);
	if (tc == NULL) {
		// NLMSG does not contain a tcmsg.
		return(FALSE);
	}

	tc->tcm_info = info;

	return(TRUE);
}

void
nl_tcmsg_print_rta_types(const struct nlmsghdr *nlmsg)
{
	const struct tcmsg *tcmsg = NULL;
	struct rtattr *rta = NULL;
	int len = 0;

	g_assert(nlmsg != NULL);

	tcmsg = nl_nlmsg_get_tcmsg_const(nlmsg);
	if (tcmsg == NULL) {
		// NLMSG does not contain a ifinfomsg.
		return;
	}

	rta = TCA_RTA(tcmsg);

	len = nl_nlmsg_get_length(nlmsg);
	len = len - NLMSG_LENGTH(NLMSG_ALIGN(sizeof(struct tcmsg)));

	while(RTA_OK(rta, len)) {
		g_print("rta->rta_type:%i\n", rta->rta_type);
		
		rta = RTA_NEXT(rta, len);
	}

	return;
}

const void*
nl_tcmsg_get_rta(const struct nlmsghdr *nlmsg,
		 unsigned short type)
{
	const struct tcmsg *tcmsg = NULL;
	struct rtattr *rta = NULL;
	int len = 0;

	g_assert(nlmsg != NULL);

	tcmsg = nl_nlmsg_get_tcmsg_const(nlmsg);
	if (tcmsg == NULL) {
		// NLMSG does not contain a ifinfomsg.
		return(NULL);
	}

	rta = TCA_RTA(tcmsg);

	len = nl_nlmsg_get_length(nlmsg);
	len = len - NLMSG_LENGTH(NLMSG_ALIGN(sizeof(struct tcmsg)));
	
	while(RTA_OK(rta, len)) {
		if (rta->rta_type == type) {
			return(rta);
		}

		rta = RTA_NEXT(rta, len);
	}

	return(NULL);
}

const void*
nl_tcmsg_get_rta_val(const struct nlmsghdr *nlmsg,
		     unsigned short type)
{
	const struct rtattr *rta = NULL;

	g_assert(nlmsg != NULL);
	
	rta = nl_tcmsg_get_rta(nlmsg, type);
	if (rta == NULL) {
		return(NULL);
	}
	
	return(RTA_DATA(rta));
}

const void*
nl_tcmsg_get_rta_sub(const struct nlmsghdr *nlmsg,
		     unsigned short parentType,
		     unsigned short type)
{
	const struct rtattr *rta = NULL, *rta2 = NULL;
	int len = 0;

	g_assert(nlmsg != NULL);

	// Get a pointer to the parent rtattr.
	rta = nl_tcmsg_get_rta(nlmsg, parentType);
	if (rta == NULL) {
		return(NULL);
	}
	len = rta->rta_len;

	// Get a pointer to the first internal rtattr.
	rta2 = RTA_DATA(rta);
	
	while (RTA_OK(rta2, len)) {
		if (rta2->rta_type == type) {
			return(rta2);
		}

		rta2 = RTA_NEXT(rta2, len);
	}
	
	return(NULL);	
}

const void*
nl_tcmsg_get_rta_sub_val(const struct nlmsghdr *nlmsg,
			 unsigned short parentType,
			 unsigned short type)
{
	const struct rtattr *rta = NULL;

	g_assert(nlmsg != NULL);
	
	rta = nl_tcmsg_get_rta_sub(nlmsg, parentType, type);
	if (rta == NULL) {
		return(NULL);
	}
	
	return(RTA_DATA(rta));
}
