/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#ifndef _NL_H_
#define _NL_H_

#include <asm/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <linux/pkt_sched.h>
#include <linux/pkt_cls.h>

#include <glib.h>

#define NL_PACKET_DATA_SIZE 4096

/*
 * A structure that represents a Netlink connection.
 */
typedef struct 
{
	int sock;
	struct sockaddr_nl addr;
} NLCon;

/*
 * A structure that represents the packets used to communicate with the kernel
 * about interfaces.
 */
typedef struct
{
	struct nlmsghdr hdr;
	struct ifinfomsg ifinfo;
	char data[NL_PACKET_DATA_SIZE];
} IfInfoPacket;

/*
 * A structure that represents the packets used to communicate with the kernel
 * about traffic control.
 */
typedef struct
{
	struct nlmsghdr hdr;
	struct tcmsg tc;
	char data[NL_PACKET_DATA_SIZE];
} TcPacket;

/* rtattr functions. */

/* depreciated function. remove soon. */
gboolean nl_rtattr_set_len(struct rtattr *rta, struct nlmsghdr *hdr);

/* IfInfoPacket functions. */
IfInfoPacket* nl_ifinfopacket_new(void);
struct nlmsghdr* nl_ifinfopacket_get_nlmsg(IfInfoPacket *pkt);
struct ifinfomsg* nl_ifinfopacket_get_ifinfomsg(IfInfoPacket *pkt);
gboolean nl_ifinfopacket_add_rtattr(IfInfoPacket *pkt, unsigned short type, unsigned short len, void *data);

/* TcPacket functions. */
TcPacket* nl_tcpacket_new(void);
struct nlmsghdr* nl_tcpacket_get_nlmsg(TcPacket *pkt);
struct tcmsg* nl_tcpacket_get_tcmsg(TcPacket *pkt);
gboolean nl_tcpacket_add_rtattr(TcPacket *pkt, unsigned short type, unsigned short len, void *data);
gboolean nl_tcpacket_add_rtattr_sub(TcPacket *pkt, unsigned short parentType, unsigned short type, unsigned short len, void *data);
gboolean nl_tcpacket_do_command(TcPacket *pkt, NLCon *nlcon);

/* Utility functions. */

/* Basic functions. */
NLCon* nl_open(unsigned groups);
size_t nl_read(NLCon *con, void *buf, int len);
gboolean nl_send(const NLCon *con, const void *buffer, const size_t len);
gboolean nl_close(NLCon *con);

/* nlmsghdr functions. */
const struct nlmsghdr* nl_get_nlmsghdr(const void *buf, const int len, const int num);
__u32 nl_nlmsg_get_length(const struct nlmsghdr *nlmsg);
__u16 nl_nlmsg_get_type(const struct nlmsghdr *nlmsg);
__u16 nl_nlmsg_get_flags(const struct nlmsghdr *nlmsg);
__u32 nl_nlmsg_get_pid(const struct nlmsghdr *nlmsg);
__u32 nl_nlmsg_get_seq(const struct nlmsghdr *nlmsg);
gboolean nl_nlmsg_set_length(struct nlmsghdr *nlmsg, __u32 length);
gboolean nl_nlmsg_set_type(struct nlmsghdr *nlmsg, __u16 type);
gboolean nl_nlmsg_set_flags(struct nlmsghdr *nlmsg, __u16 flags);
gboolean nl_nlmsg_set_pid(struct nlmsghdr *nlmsg, __u32 pid);
gboolean nl_nlmsg_set_seq(struct nlmsghdr *nlmsg, __u32 seq);
const struct ifinfomsg* nl_nlmsg_get_ifinfomsg_const(const struct nlmsghdr *nlmsg);
struct ifinfomsg* nl_nlmsg_get_ifinfomsg(struct nlmsghdr *nlmsg);
const struct tcmsg* nl_nlmsg_get_tcmsg_const(const struct nlmsghdr *nlmsg);
struct tcmsg* nl_nlmsg_get_tcmsg(struct nlmsghdr *nlmsg);
const struct rtmsg* nl_nlmsg_get_rtmsg(const struct nlmsghdr *nlmsg);
gboolean nl_nlmsg_is_done(const struct nlmsghdr *nlmsg);
gboolean nl_nlmsg_is_error(const struct nlmsghdr *nlmsg);
gboolean nl_nlmsg_is_multi(const struct nlmsghdr *nlmsg);

/* ifinfomsg functions. */
gboolean nl_ifinfo_verify(const struct nlmsghdr *nlmsg);
unsigned char nl_ifinfo_get_family(const struct nlmsghdr *nlmsg);
unsigned short nl_ifinfo_get_type(const struct nlmsghdr *nlmsg);
int nl_ifinfo_get_index(const struct nlmsghdr *nlmsg);
unsigned nl_ifinfo_get_flags(const struct nlmsghdr *nlmsg);
unsigned nl_ifinfo_get_change(const struct nlmsghdr *nlmsg);
gboolean nl_ifinfo_set_family(struct nlmsghdr *nlmsg, unsigned char family);
gboolean nl_ifinfo_set_type(struct nlmsghdr *nlmsg, unsigned short type);
gboolean nl_ifinfo_set_index(struct nlmsghdr *nlmsg, int index);
gboolean nl_ifinfo_set_flags(struct nlmsghdr *nlmsg, unsigned flags);
gboolean nl_ifinfo_set_change(struct nlmsghdr *nlmsg, unsigned change);
const void* nl_ifinfo_get_rta_val(const struct nlmsghdr *nlmsg, unsigned short type);

/* tcmsg functions. */
gboolean nl_tcmsg_verify(const struct nlmsghdr *nlmsg);
unsigned char nl_tcmsg_get_family(const struct nlmsghdr *nlmsg);
int nl_tcmsg_get_ifindex(const struct nlmsghdr *nlmsg);
__u32 nl_tcmsg_get_handle(const struct nlmsghdr *nlmsg);
__u32 nl_tcmsg_get_parent(const struct nlmsghdr *nlmsg);
__u32 nl_tcmsg_get_info(const struct nlmsghdr *nlmsg);
gboolean nl_tcmsg_set_family(struct nlmsghdr *nlmsg, unsigned char family);
gboolean nl_tcmsg_set_ifindex(struct nlmsghdr *nlmsg, int index);
gboolean nl_tcmsg_set_handle(struct nlmsghdr *nlmsg, __u32 handle);
gboolean nl_tcmsg_set_parent(struct nlmsghdr *nlmsg, __u32 parent);
gboolean nl_tcmsg_set_info(struct nlmsghdr *nlmsg, __u32 info);
void nl_tcmsg_print_rta_types(const struct nlmsghdr *nlmsg);
const void* nl_tcmsg_get_rta(const struct nlmsghdr *nlmsg, unsigned short type);
const void* nl_tcmsg_get_rta_val(const struct nlmsghdr *nlmsg, unsigned short type);
const void* nl_tcmsg_get_rta_sub(const struct nlmsghdr *nlmsg, unsigned short parentType, unsigned short type);
const void* nl_tcmsg_get_rta_sub_val(const struct nlmsghdr *nlmsg, unsigned short parentType, unsigned short type);

/* rtmsg functions. */
unsigned char nl_rtmsg_get_family(const struct nlmsghdr *nlmsg);

#endif
